/** 结构体模拟解析-支持一维二维数组，请不要定义三维数组 */
export default class CStruct 
{
    /**1字节 */ static byte1: number = 1;
    /**2字节 */ static byte2: number = 2;
    /**4字节 */ static byte4: number = 4;
    /**8字节 */ static byte8: number = 8;
    /**uint32*/ static BITS_32:number = 4294967296; // 2 ** 32,

    /** 反序列化数据 */
    static init(root_obj, dataBuffer:any[], dataIndex: number = 0) 
    {
        let self = root_obj;

        //属性数组
        let property = Object.getOwnPropertyNames(self);
        let nDataIndex = dataIndex;
        let nDataCount = dataIndex;

        //解析完成
        let data_len = dataBuffer.length;
        if (data_len == 0 || data_len < nDataIndex) {
            cc.warn("decode is doned. dataBuffer len:" + data_len + ",dataIndex:" + nDataIndex);
            return;
        }

        //遍历属性
        for (let i=0, c=property.length; i<c; i++) 
        {
            //解码提前结束
            if (data_len <= nDataCount) {
                cc.warn(" decode is doned. dataBuffer len:" + data_len + ",dataIndex:" + nDataCount);
                break;
            }

            let key = property[i];
            let key_size = this.getKeySize(key);
            let type = typeof self[key];

            /**对象|数组 */
            if (type == "object") 
            {
                let p_obj = self[key];

                //对象类型
                if( key_size==0 )
                {
                    //对象数组
                    let desc_key = key + '_desc';
                    if( self[desc_key] )
                    {
                        let desc = self[desc_key]();
                        let array_obj_size = this.initArrayObject(p_obj, desc, dataBuffer, nDataCount);
                        nDataCount += array_obj_size;
                    }
                    else
                    {
                        this.init(p_obj, dataBuffer, nDataCount);
                        let obj_size = this.getDataSize(p_obj);
                        nDataCount += obj_size;
                    }
                }
                //数组类型
                else
                {
                    let arraySize = this.initArrayNumber(root_obj, key, dataBuffer, nDataCount);
                    nDataCount += arraySize;
                }
            }
            else if (type == "undefined") {
                cc.warn("struct.undefined.key." + key);
            }
            else 
            {
                //解码
                let value_size = this.getKeySize(key);
                self[key] = this.getNumber(dataBuffer, value_size, nDataCount);
                nDataCount += value_size;
            }
        }
        return nDataCount;
    }

    /**初始化对象数组、最多支持二维对象数组 */
    static initArrayObject(p_obj:any, obj_desc:any, dataBuffer:any[], dataIndex:number)
    {
        let obj_type = obj_desc['type'];
        let obj_tp = new obj_type();
        let obj_size = this.getDataSize(obj_tp);
        let nDataCount = 0;
        let dataSize = dataBuffer.length;
        let subType = typeof p_obj[0];
        for(let j=0, k=p_obj.length; j<k; ++j)
        {
            if( dataSize-(nDataCount+dataIndex)<obj_size && obj_size!=0 ){
                cc.warn( "dataBuffer is too small! decode array end.dataSize:" + dataSize );
                break;
            }

            /**子项为数组 */
            if( subType=="object" )
            {
                let ary_obj_count = this.initArrayObject(p_obj[j], obj_desc, dataBuffer, nDataCount+dataIndex);
                nDataCount += ary_obj_count;
            }
            /**子项为对象 */
            else
            {
                let obj_new = new obj_type();
                if(obj_size==0) obj_size = this.getDataSize(obj_new);
                this.init(obj_new, dataBuffer, nDataCount + dataIndex );
                nDataCount += obj_size;
                p_obj[j] = obj_new;
            }
        }
        return nDataCount;
    }

    /**初始化数组，最多支持二维数组 */
    static initArrayNumber(root_obj:any[], key:string, dataBuffer:any[], dataIndex=0)
    {
        let p_obj = root_obj[key];
        let key_size = this.getKeySize(key);
        let obj_len = p_obj.length;
        let nDataCount = 0;
        let dataSize = dataBuffer.length;

        let subType = typeof p_obj[0];

        for(let j=0; j<obj_len; ++j)
        {
            if( dataSize-(nDataCount+dataIndex)<key_size ){
                cc.warn( "dataBuffer is too small! decode array end.dataSize:" + dataSize );
                break;
            }

            /**子项为数组 */
            if( subType=="object" )
            {
                p_obj[j] = this.getArrayNumber(p_obj[j], key_size, dataBuffer, nDataCount+dataIndex);
                nDataCount += key_size * p_obj[j].length;
            }
            /**子项为基础类型 */
            else
            {
                p_obj[j] = this.getNumber(dataBuffer, key_size, nDataCount+dataIndex);
                nDataCount += key_size;
            }
        }
        return nDataCount;
    }
    
    /**序列化大小 */
    static getDataSize(root_obj: Object): number 
    {
        if( root_obj==0 ) return 0;
        let self = root_obj;
        let root_size = 0;

        //属性数组
        let property = Object.getOwnPropertyNames(self);

        //遍历属性
        for (let i=0,len=property.length; i<len; i++) 
        {
            let key = property[i];
            let key_size = this.getKeySize(key);
            let type = typeof self[key];

            //对象 | 数组
            if (type == "object") 
            {
                let p_obj = self[key];

                //数组-对象类型
                if( key_size==0 )
                {
                    let desc_key = key + '_desc';

                    /**数组对象 */
                    if( self[desc_key] )
                    {
                        let desc = self[desc_key]();
                        let obj_new = new desc['type'];
                        let obj_size = this.getDataSize(obj_new);
                        root_size = root_size + obj_size * p_obj.length;
                    }
                    else
                    {
                        /**单个对象 */
                        let obj_size = this.getDataSize(p_obj);
                        root_size = root_size + obj_size;
                    }
                }
                //数组-基础类型
                else
                {
                    //二维类型
                    if(typeof p_obj[0]=='object'){
                        root_size = root_size + key_size * p_obj.length * p_obj[0].length;
                    }
                    else{
                        root_size = root_size + key_size * p_obj.length;
                    }
                }
            }else if (type == "undefined") {
                cc.warn("struct.undefined.key." + key);
            }
            else 
            {
                root_size = root_size + key_size;
            }
        }
        return root_size;
    }

    /**序列化字节数据 */
    static getDataBytes(root_obj: Object) 
    {
        let self = root_obj;
        let dataBuffer = [];

        let property = Object.getOwnPropertyNames(self);

        for (let i=0, c=property.length; i<c; ++i) 
        {
            let key = property[i];
            let key_size = this.getKeySize(key);
            let type = typeof self[key];

            /**对象|数组 */
            if (type == "object")
            {
                let p_obj = self[key];
                
                /**子项为对象*/
                if( key_size==0 )
                {
                    let desc_key = key + '_desc';

                    /**子项为 对象数组 */
                    if( self[desc_key] )
                    {
                        let desc = self[desc_key]();
                        let ary_obj_bytes = this.getArrayObjectBytes(p_obj, desc);
                        dataBuffer = dataBuffer.concat(ary_obj_bytes);
                    }
                    /**子项为对象 */
                    else
                    {
                        let e_bytes = this.getDataBytes(p_obj);
                        dataBuffer = dataBuffer.concat(e_bytes);
                    }
                }
                /**子项为基本数据*/
                else
                {
                    let array_bytes = this.getArrayBytes(p_obj, key_size);
                    dataBuffer = dataBuffer.concat(array_bytes);
                }
            }
            else if (type == "undefined") {
                cc.warn("struct.undefined.key." + key);
            }
            else {
                let size = this.getKeySize(key);
                dataBuffer = dataBuffer.concat(this.getValueBytes(self[key], size));
            }
        }

        return dataBuffer;
    }

    /** 分析关键字字节长度 */
    static getKeySize(key: string): number {
        let key_0 = key.substr(0, 1);
        let key_1 = key.substr(0, 2);

        if (key_0 == 'c') return this.byte1; //byte
        if (key_0 == 'b') return this.byte1; //bool
        else if (key_1 == 'cb') return this.byte1; //byte
        else if (key_1 == 'by') return this.byte1; //byte

        else if (key_1 == 'sz') return this.byte2; //w_char
        else if (key_0 == 'w') return this.byte2; //WORD

        else if (key_1 == 'dw') return this.byte4; //DWORD
        else if (key_0 == 'i') return this.byte4; //int
        else if (key_0 == 'n') return this.byte4; //int

        else if (key_1 == 'll') return this.byte8; //longlong
        else if (key_0 == 'l') return this.byte8; //longlong
        return 0;
    }

    /**变量转换字节 */
    static getValueBytes(tp_value: number, tp_value_size: number): any[] 
    {
        //正常取位 低位在前 高位在后
        // var ch = tp_value, st = [];
        // if( tp_value_size<=4 )
        // {
        //     for (var i = 0; i < tp_value_size; i++ )
        //     {  
        //         let byte = (ch & 0xFF); 
        //         st.push(byte);

        //         //shift value down by 1 byte
        //         ch = ch >> 8;          
        //     }
        // }
        // else
        // {
        //     //符号位
        //     var cbSigned = 0;
        //     var lTempNumber = tp_value;
        //     if( tp_value<0 ) {
        //         cbSigned=1;
        //         lTempNumber = tp_value*-1;
        //     }
        //     var lBaseValue = Math.pow(256, tp_value_size-1);
        //     var tempValue = 0;

        //     //先取高位
        //     for(let i=0; i<tp_value_size; i++)
        //     {
        //         tempValue = Math.floor(lTempNumber/lBaseValue);

        //         if( tempValue>0 && lTempNumber>=lBaseValue ){
        //             st.push( tempValue&0xFF );
        //         }
        //         else{
        //             st.push(0);
        //             tempValue=0;
        //         }

        //         lTempNumber -= ( tempValue * lBaseValue );
        //         lBaseValue = lBaseValue/256;
        //     }

        //     //符号位补码
        //     if( cbSigned )
        //     {
        //         for(let i=0; i<tp_value_size; i++)
        //         {
        //             st[i] = ~st[i];
        //             st[i] &= 0xff; 
        //         }
        //     }

        //     //反序排列
        //     st.reverse();
        // }

        // //常规排序
        // return st;

        /*low-32位 正常取位 低位在前 高位在后 */
        let value = tp_value ? tp_value : 0;
        let value_size = tp_value_size;
        let top = Math.floor(value / this.BITS_32);
        let low = value - top * this.BITS_32;
        
        let ch = value, st = [];
        let low_size = value_size;
        let top_size = value_size - 4;
        if(top_size>0){
            ch = low;
            low_size = 4;
        }

        /**low 32位 */
        for (let i = 0; i < low_size; i++) 
        {
            let byte = (ch & 0xFF);
            st.push(byte);
            ch = ch >> 8;
        }

        /**top-32位 */
        if(top_size>0)
        {
            ch = top;
            for (let i = 0; i < top_size; i++) 
            {
                let byte = (ch & 0xFF);
                st.push(byte);
                ch = ch >> 8;
            }
        }

        //常规排序
        return st;
    }

    static getArrayObjectBytes(p_obj:any, desc:any)
    {
        let dataBuffer = [];
        if(p_obj.length==0) return dataBuffer;

        let obj_type = desc['type'];
        let subType = typeof p_obj[0];
        for(let j=0, k=p_obj.length; j<k; ++j )
        {
            if( subType=="object" )
            {
                let sub_ary_bytes = this.getArrayObjectBytes(p_obj[j], desc);
                dataBuffer = dataBuffer.concat(sub_ary_bytes);
            }
            else
            {
                let element = p_obj[j];
                if(!element) element = new obj_type();
                let e_bytes = this.getDataBytes(element);
                dataBuffer = dataBuffer.concat(e_bytes);
            }
        }
        return dataBuffer;
    }

    /**数组转换字节 */
    static getArrayBytes(p_obj:any[], key_size:number)
    {
        let dataBuffer = []
        for(let j=0, k=p_obj.length; j<k; ++j )
        {
            let element = p_obj[j];
            if(!element)element=0;
            let e_bytes = this.getValueBytes(element, key_size);
            dataBuffer = dataBuffer.concat(e_bytes);
        };
        return dataBuffer;
    }

    /**字节转化数组变量 */
    static getArrayNumber(p_obj:any[], key_size:number, dataBuffer:any[], data_offset=0)
    {
        let obj_len = p_obj.length;
        let nDataCount = data_offset;
        let dataSize = dataBuffer.length;
        for(let j=0; j<obj_len; ++j)
        {
            if( dataSize-nDataCount<key_size ){
                cc.warn( "dataBuffer is too small! decode array end.dataSize:" + dataSize );
                break;
            }
            p_obj[j] = this.getNumber(dataBuffer, key_size, nDataCount);
            nDataCount += key_size;
        }
        return p_obj;
    }

    /**字节转变量*/
    static getNumber(dataBuffer:any[], value_size:number, dataIndex:number=0) 
    {
        let cbTemp:number = 0;
        let lTempNumber:number = 0;
        let wValueSize:number = value_size;
    
        let cbSigned:number = 0;
        let lBaseValue:number = 1;
    
        if( dataBuffer.length<dataIndex+wValueSize )
        {
            cc.warn( "dataBuffer is too small:" + dataBuffer.length + ",dataIndex:" + (dataIndex+wValueSize) );
            return;
        }
    
        //常规数值
        if( wValueSize<=4 )
        {
            for(let i=0; i<wValueSize; i++)
            {
                cbTemp = dataBuffer[dataIndex+i]; 
    
                lTempNumber |= (cbTemp<<i*8);                                         
            }
    
            //无符号数
            return lTempNumber;
        }
        else
        {
            for(let i=0; i<wValueSize; i++)
            {
                cbTemp = dataBuffer[dataIndex+i];   
    
                //取符号位
                if( i===0 ){
                    var cbHigh = dataBuffer[dataIndex+i+wValueSize-1];
                    cbSigned = cbHigh & 0x80;
                }
    
                //进制数
                if( i>0 ){
                    lBaseValue = lBaseValue*256;
                }
    
                //有符号数
                if( cbSigned==0x80 ){
                    cbTemp = ~cbTemp;
                    cbTemp &= 0xff; 
                }
    
                lTempNumber += ( cbTemp * lBaseValue );
            }
    
            //有符号数
            if( cbSigned==0x80 ) return - ( lTempNumber+1 );
    
            //无符号数
            return lTempNumber;
        }
    }

    static sizeof(obj_class:any)
    {
        let item = new obj_class();
        return this.getDataSize(item);
    }

    /**获取多结构数据拼接 */
    static getMoreDataBytes(...arg)
    {
        let dataBuffer = [];
        for(let i=0,c=arguments.length; i<c; i++)
        {
            let tp_bytes = CStruct.getDataBytes(arguments[i]);
            dataBuffer = dataBuffer.concat(tp_bytes);
        }
        return dataBuffer;
    }
}

/** Array.from 方法查漏 */
if (!Array.from) {
    Array.from = (function () {
        var toStr = Object.prototype.toString;
        var isCallable = function (fn) {
            return typeof fn === 'function' || toStr.call(fn) === '[object Function]';
        };
        var toInteger = function (value) {
            var number = Number(value);
            if (isNaN(number)) { return 0; }
            if (number === 0 || !isFinite(number)) { return number; }
            return (number > 0 ? 1 : -1) * Math.floor(Math.abs(number));
        };
        var maxSafeInteger = Math.pow(2, 53) - 1;
        var toLength = function (value) {
            var len = toInteger(value);
            return Math.min(Math.max(len, 0), maxSafeInteger);
        };

        // The length property of the from method is 1.
        return function from(arrayLike/*, mapFn, thisArg */) {
            // 1. Let C be the this value.
            var C = this;

            // 2. Let items be ToObject(arrayLike).
            var items = Object(arrayLike);

            // 3. ReturnIfAbrupt(items).
            if (arrayLike == null) {
                throw new TypeError("Array.from requires an array-like object - not null or undefined");
            }

            // 4. If mapfn is undefined, then let mapping be false.
            var mapFn = arguments.length > 1 ? arguments[1] : void undefined;
            var T;
            if (typeof mapFn !== 'undefined') {
                // 5. else      
                // 5. a If IsCallable(mapfn) is false, throw a TypeError exception.
                if (!isCallable(mapFn)) {
                    throw new TypeError('Array.from: when provided, the second argument must be a function');
                }

                // 5. b. If thisArg was supplied, let T be thisArg; else let T be undefined.
                if (arguments.length > 2) {
                    T = arguments[2];
                }
            }

            // 10. Let lenValue be Get(items, "length").
            // 11. Let len be ToLength(lenValue).
            var len = toLength(items.length);

            // 13. If IsConstructor(C) is true, then
            // 13. a. Let A be the result of calling the [[Construct]] internal method 
            // of C with an argument list containing the single item len.
            // 14. a. Else, Let A be ArrayCreate(len).
            var A = isCallable(C) ? Object(new C(len)) : new Array(len);

            // 16. Let k be 0.
            var k = 0;
            // 17. Repeat, while k < len… (also steps a - h)
            var kValue;
            while (k < len) {
                kValue = items[k];
                if (mapFn) {
                    A[k] = typeof T === 'undefined' ? mapFn(kValue, k) : mapFn.call(T, kValue, k);
                } else {
                    A[k] = kValue;
                }
                k += 1;
            }
            // 18. Let putStatus be Put(A, "length", len, true).
            A.length = len;
            // 20. Return A.
            return A;
        };
    }());
}